package netmap

import (
	"git.frostfs.info/TrueCloudLab/frostfs-node/internal/logs"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/innerring/processors/governance"
	cntClient "git.frostfs.info/TrueCloudLab/frostfs-node/pkg/morph/client/container"
	netmapEvent "git.frostfs.info/TrueCloudLab/frostfs-node/pkg/morph/event/netmap"
	"go.uber.org/zap"
)

// Process new epoch notification by setting global epoch value and resetting
// local epoch timer.
func (np *Processor) processNewEpoch(ev netmapEvent.NewEpoch) {
	epoch := ev.EpochNumber()

	epochDuration, err := np.netmapClient.EpochDuration()
	if err != nil {
		np.log.Warn(logs.NetmapCantGetEpochDuration,
			zap.String("error", err.Error()))
	} else {
		np.epochState.SetEpochDuration(epochDuration)
	}

	np.epochState.SetEpochCounter(epoch)

	h, err := np.netmapClient.MorphTxHeight(ev.TxHash())
	if err != nil {
		np.log.Warn(logs.NetmapCantGetTransactionHeight,
			zap.String("hash", ev.TxHash().StringLE()),
			zap.String("error", err.Error()))
	}

	if err := np.epochTimer.ResetEpochTimer(h); err != nil {
		np.log.Warn(logs.NetmapCantResetEpochTimer,
			zap.String("error", err.Error()))
	}

	// get new netmap snapshot
	networkMap, err := np.netmapClient.NetMap()
	if err != nil {
		np.log.Warn(logs.NetmapCantGetNetmapSnapshotToPerformCleanup,
			zap.String("error", err.Error()))

		return
	}

	prm := cntClient.StartEstimationPrm{}

	prm.SetEpoch(epoch - 1)
	prm.SetHash(ev.TxHash())

	if epoch > 0 && np.alphabetState.IsAlphabet() { // estimates are invalid in genesis epoch
		err = np.containerWrp.StartEstimation(prm)

		if err != nil {
			np.log.Warn(logs.NetmapCantStartContainerSizeEstimation,
				zap.Uint64("epoch", epoch),
				zap.String("error", err.Error()))
		}
	}

	np.netmapSnapshot.update(*networkMap, epoch)
	np.handleCleanupTick(netmapCleanupTick{epoch: epoch, txHash: ev.TxHash()})
	np.handleAlphabetSync(governance.NewSyncEvent(ev.TxHash()))
	np.handleNotaryDeposit(ev)
}

// Process new epoch tick by invoking new epoch method in network map contract.
func (np *Processor) processNewEpochTick() {
	if !np.alphabetState.IsAlphabet() {
		np.log.Info(logs.NetmapNonAlphabetModeIgnoreNewEpochTick)
		return
	}

	nextEpoch := np.epochState.EpochCounter() + 1
	np.log.Debug(logs.NetmapNextEpoch, zap.Uint64("value", nextEpoch))

	err := np.netmapClient.NewEpoch(nextEpoch, false)
	if err != nil {
		np.log.Error(logs.NetmapCantInvokeNetmapNewEpoch, zap.Error(err))
	}
}
