package putsvc

import (
	"context"
	"fmt"

	objectCore "git.frostfs.info/TrueCloudLab/frostfs-node/pkg/core/object"
	"git.frostfs.info/TrueCloudLab/frostfs-sdk-go/object"
	oid "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/object/id"
	"git.frostfs.info/TrueCloudLab/frostfs-sdk-go/object/transformer"
)

// ObjectStorage is an object storage interface.
type ObjectStorage interface {
	// Put must save passed object
	// and return any appeared error.
	Put(context.Context, *object.Object) error
	// Delete must delete passed objects
	// and return any appeared error.
	Delete(ctx context.Context, tombstone oid.Address, toDelete []oid.ID) error
	// Lock must lock passed objects
	// and return any appeared error.
	Lock(ctx context.Context, locker oid.Address, toLock []oid.ID) error
	// IsLocked must clarify object's lock status.
	IsLocked(context.Context, oid.Address) (bool, error)
}

type localTarget struct {
	storage ObjectStorage

	obj  *object.Object
	meta objectCore.ContentMeta
}

func (t *localTarget) WriteObject(obj *object.Object, meta objectCore.ContentMeta) error {
	t.obj = obj
	t.meta = meta

	return nil
}

func (t *localTarget) Close(ctx context.Context) (*transformer.AccessIdentifiers, error) {
	switch t.meta.Type() {
	case object.TypeTombstone:
		err := t.storage.Delete(ctx, objectCore.AddressOf(t.obj), t.meta.Objects())
		if err != nil {
			return nil, fmt.Errorf("could not delete objects from tombstone locally: %w", err)
		}
	case object.TypeLock:
		err := t.storage.Lock(ctx, objectCore.AddressOf(t.obj), t.meta.Objects())
		if err != nil {
			return nil, fmt.Errorf("could not lock object from lock objects locally: %w", err)
		}
	default:
		// objects that do not change meta storage
	}

	if err := t.storage.Put(ctx, t.obj); err != nil { //TODO
		return nil, fmt.Errorf("(%T) could not put object to local storage: %w", t, err)
	}

	id, _ := t.obj.ID()

	return &transformer.AccessIdentifiers{
		SelfID: id,
	}, nil
}
